/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor;

import java.io.IOException;
import java.util.Deque;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Filer;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.TypeElement;
import javax.tools.FileObject;
import javax.tools.StandardLocation;

import com.is2t.giml.annotation.processor.ant.GIMLAnnotationProcessorTask;
import com.is2t.giml.annotation.processor.generator.DescriptionSourceGenerator;
import com.is2t.giml.annotation.processor.model.Attribute;
import com.is2t.giml.annotation.processor.model.Child;
import com.is2t.giml.annotation.processor.model.Description;

/**
 *
 */
@SupportedAnnotationTypes({ "ej.giml.annotation.Element" })
@SupportedSourceVersion(SourceVersion.RELEASE_7)
public class GIMLAnnotationProcessor extends AbstractProcessor {

	private static final Logger LOGGER = Logger.getLogger(GIMLAnnotationProcessor.class.getName());

	@Override
	public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
		try {
			Filer filer = this.processingEnv.getFiler();
			DescriptionSourceGenerator generator = new DescriptionSourceGenerator(filer);
			Set<? extends javax.lang.model.element.Element> rootElements = roundEnv.getRootElements();
			for (javax.lang.model.element.Element element : rootElements) {
				if (element.getKind() == ElementKind.CLASS) {
					GIMLElementVisitor visitor = new GIMLElementVisitor();
					element.accept(visitor, Boolean.TRUE);
					Description elementDescription = visitor.getDescription();
					if (elementDescription != null) {
						// Get icons.
						String tag = elementDescription.getName().toLowerCase();
						elementDescription.setIcon16(getIcon(filer, element, tag + "_16x16.png"));
						elementDescription.setIcon24(getIcon(filer, element, tag + "_24x24.png"));

						generator.createDescriptionSource(elementDescription);
						// now log found description (useful when remote building artifact)
						this.logElementDescription(elementDescription);
					}
				}
			}
			generator.createDescriptionPackageSource();
		} catch (Exception e) {
			throw new RuntimeException(e.getMessage(), e);
		}

		if (!GIMLAnnotationProcessorTask.DEBUG_MODE) {
			// FIXME @see WI 16719
			System.exit(0);
		}

		return true;
	}

	private FileObject getIcon(Filer filer, javax.lang.model.element.Element element, String name) {
		try {
			return filer.getResource(StandardLocation.SOURCE_PATH, element.getEnclosingElement().toString(),
					"doc-files/" + name);
		} catch (IOException e) {
			// The icon does not exist.
		}
		return null;
	}

	private void logElementDescription(Description elementDescription) {
		LOGGER.info(String.format("Found element '%s' (class=%s,isRoot=%s)", elementDescription.getName(),
				elementDescription.getElementClassName(), elementDescription.isRoot()));
		List<Attribute> elementAttributes = elementDescription.getAttributes();
		if (!elementAttributes.isEmpty()) {
			LOGGER.info(" > Element attributes:");
			for (Attribute attribute : elementAttributes) {
				LOGGER.info(String.format("\t'%s' (class=%s,required=%s,default=%s)", attribute.getName(),
						attribute.getAttributeClassName(), attribute.isRequired(), attribute.getDefaultValue()));
			}
		}
		List<Child> elementChildren = elementDescription.getChildren();
		if (!elementChildren.isEmpty()) {
			LOGGER.info(" > Element children:");
			for (Child child : elementChildren) {
				LOGGER.info(String.format("\t'%s' (prefix=%s,cardinality=%s,ordered=%s)", child.getName(),
						child.getPrefix(), child.getCardinality(), child.isOrdered()));
				Deque<Attribute> childAttributes = child.getAttributes();
				if (!childAttributes.isEmpty()) {
					LOGGER.info("\t > Child attributes:");
					for (Attribute attribute : childAttributes) {
						LOGGER.info(String.format("\t\t'%s' (class=%s,required=%s,default=%s)", attribute.getName(),
								attribute.getAttributeClassName(), attribute.isRequired(),
								attribute.getDefaultValue()));
					}
				}
			}
		}

	}

}
