/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.generator;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import javax.annotation.processing.Filer;
import javax.tools.FileObject;
import javax.tools.JavaFileObject;
import javax.tools.StandardLocation;

import com.is2t.giml.annotation.processor.model.Description;
import com.is2t.giml.description.DescriptionPackage;

/**
 * The description source generator allows to generate Java source content. Generated classes are related to
 * {@link com.is2t.giml.description.Description} framework.
 */
public class DescriptionSourceGenerator {

	private static final String SERVICES_PACKAGE_NAME = "META-INF/services";
	private static final String GENERATED_PACKAGE_NAME = "com.is2t.giml.description.generated";
	private static final String DESCRIPTION_PACKAGE_CLASS_NAME = "GeneratedDescriptionPackage";

	private final Filer _filer; // NOSONAR
	private final TemplateHelper _st; // NOSONAR
	private final List<String> _descriptionClassNames; // NOSONAR

	/**
	 * Instantiates a new generator that will use given filer to create Java source content.
	 *
	 * @throws IllegalArgumentException
	 *             if class names cannot be dynamically read from {@link ClassName} interface.
	 * @throws IllegalAccessException
	 *             if class names cannot be dynamically read from {@link ClassName} interface.
	 */
	public DescriptionSourceGenerator(Filer filer) throws IllegalArgumentException, IllegalAccessException {
		this._filer = filer;
		this._st = new TemplateHelper();
		this._descriptionClassNames = new ArrayList<String>();
	}

	/**
	 * Generates Java source file of a class that implements {@link com.is2t.giml.description.Description}.
	 *
	 * @param description
	 *            the model to use to generate class content.
	 * @throws IOException
	 *             if source file cannot be created.
	 */
	public void createDescriptionSource(Description description) throws IOException {
		this._descriptionClassNames.add(this.getClassName(description));
		JavaFileObject sourceFile = this._filer.createSourceFile(getFullyQualifiedName(description));
		try (PrintWriter writer = new PrintWriter(sourceFile.openWriter())) {
			String className = this.getClassName(description);
			writer.print(this._st.generateClassDescription(description, GENERATED_PACKAGE_NAME, className));
		}

		// Copy icons if exist.
		String tag = description.getName().toLowerCase();
		FileObject icon16 = description.getIcon16();
		if (icon16 != null) {
			FileObject output = this._filer.createResource(StandardLocation.SOURCE_OUTPUT, GENERATED_PACKAGE_NAME,
					tag + "_16x16.png");
			copy(icon16, output);
		}
		FileObject icon24 = description.getIcon24();
		if (icon24 != null) {
			FileObject output = this._filer.createResource(StandardLocation.SOURCE_OUTPUT, GENERATED_PACKAGE_NAME,
					tag + "_24x24.png");
			copy(icon24, output);
		}

	}

	private void copy(FileObject input, FileObject output) throws IOException {
		InputStream reader = input.openInputStream();
		OutputStream writer = output.openOutputStream();
		byte[] buffer = new byte[1024];
		while (true) {
			int length = reader.read(buffer);
			if (length == -1) {
				break;
			}
			writer.write(buffer, 0, length);
		}
	}

	/**
	 * Generates Java source file of a class that implements {@link DescriptionPackage}. All previously generates
	 * {@link com.is2t.giml.description.Description} classes will be added to
	 * {@link DescriptionPackage#loadDescriptions()} method.
	 *
	 * In addition, service declaration is generated in {@value #SERVICES_PACKAGE_NAME} folder.
	 *
	 * @throws IOException
	 *             if source file or service declaration cannot be created.
	 */
	public void createDescriptionPackageSource() throws IOException {
		if (this._descriptionClassNames.size() > 0) {
			// Generate description package class
			JavaFileObject sourceFile = this._filer
					.createSourceFile(this.getFullyQualifiedName(DESCRIPTION_PACKAGE_CLASS_NAME));
			try (PrintWriter writer = new PrintWriter(sourceFile.openWriter())) {
				writer.print(this._st.generateClassDescriptionPackage(this._descriptionClassNames,
						GENERATED_PACKAGE_NAME, DESCRIPTION_PACKAGE_CLASS_NAME));
			}

			// Generate description package service declaration
			FileObject serviceFile = this._filer.createResource(StandardLocation.SOURCE_OUTPUT, "",
					SERVICES_PACKAGE_NAME + "/" + ClassName.GIML_DESCRIPTIONPACKAGE);
			try (PrintWriter writer = new PrintWriter(serviceFile.openWriter())) {
				writer.print(this._st.generateServiceDescriptionPackage(GENERATED_PACKAGE_NAME,
						DESCRIPTION_PACKAGE_CLASS_NAME));
			}
		}
	}

	private String getFullyQualifiedName(Description description) {
		return this.getFullyQualifiedName(this.getClassName(description));
	}

	private String getFullyQualifiedName(String className) {
		return GENERATED_PACKAGE_NAME.concat(".").concat(className);
	}

	private String getClassName(Description description) {
		String name = description.getName();
		name = name.replaceAll("\\s", "");
		name = name.replaceAll("[\\-_]", "");
		name = name.substring(0, 1).toUpperCase() + name.substring(1);
		return name + "Description";
	}

}
