/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.cssgenerator.selector;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.w3c.css.sac.Condition;
import org.w3c.css.sac.Selector;

/**
 *
 */
public class SelectorFactory {

	private static final Logger LOGGER = Logger.getLogger(SelectorFactory.class.getSimpleName());

	private static final List<com.is2t.giml.cssgenerator.selector.Selector> CONDITIONAL_SELECTORS = new ArrayList<>();
	private static final List<com.is2t.giml.cssgenerator.selector.Selector> SIBLING_SELECTORS = new ArrayList<>();
	private static final List<com.is2t.giml.cssgenerator.selector.Selector> DESCENDANT_SELECTORS = new ArrayList<>();
	private static final List<com.is2t.giml.cssgenerator.selector.Selector> ELEMENT_SELECTORS = new ArrayList<>();

	static {
		initialize();
	}

	public static void initialize() {
		CONDITIONAL_SELECTORS.clear();
		SIBLING_SELECTORS.clear();
		DESCENDANT_SELECTORS.clear();
		ELEMENT_SELECTORS.clear();

		CONDITIONAL_SELECTORS.add(new AndCombinatorSelector());
		CONDITIONAL_SELECTORS.add(new AttributeSetSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValueContainSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValueEndSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValueListSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValuePrefixSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValueSelector());
		CONDITIONAL_SELECTORS.add(new AttributeValueStartSelector());
		CONDITIONAL_SELECTORS.add(new IdSelector());
		CONDITIONAL_SELECTORS.add(new ClassSelector());
		for (String pseudoSelector : PseudoClassSelector.getAllPseudoClasses()) {
			CONDITIONAL_SELECTORS.add(new PseudoClassSelector(pseudoSelector));
		}

		SIBLING_SELECTORS.add(new AdjacentSiblingCombinator());
		SIBLING_SELECTORS.add(new GeneralSiblingCombinator());

		DESCENDANT_SELECTORS.add(new ChildCombinator());
		DESCENDANT_SELECTORS.add(new DescendantCombinator());

		ELEMENT_SELECTORS.add(new TypeSelector());
		ELEMENT_SELECTORS.add(new UniversalSelector());
	}

	private SelectorFactory() {
		// factory class
	}

	public static final String create(Selector selector, StringWriter writer) {
		switch (selector.getSelectorType()) {
		case Selector.SAC_CONDITIONAL_SELECTOR:
			return create(selector, writer, CONDITIONAL_SELECTORS);
		case Selector.SAC_DIRECT_ADJACENT_SELECTOR:
			return create(selector, writer, SIBLING_SELECTORS);
		case Selector.SAC_CHILD_SELECTOR:
		case Selector.SAC_DESCENDANT_SELECTOR:
			return create(selector, writer, DESCENDANT_SELECTORS);
		case Selector.SAC_ELEMENT_NODE_SELECTOR:
			return create(selector, writer, ELEMENT_SELECTORS);
		case Selector.SAC_ANY_NODE_SELECTOR:
			// FIXME GeneralAdjacentSelectorImpl returns SAC_ANY_NODE_SELECTOR as type, but it should return another
			// type...
			return create(selector, writer, SIBLING_SELECTORS);
		default:
			throw new RuntimeException("Not handled: " + selector + " > " + selector.getClass().getSimpleName());
		}
	}

	public static final String create(Condition condition, StringWriter writer) {
		String selectorVar = null;
		com.is2t.giml.cssgenerator.selector.Selector firstMatchedSelector = null;
		for (com.is2t.giml.cssgenerator.selector.Selector knownSelector : CONDITIONAL_SELECTORS) {
			if (knownSelector.matches(condition)) {
				if (selectorVar != null) {
					throw new RuntimeException(String.format(
							"Condition '%s' matches multiple condition definitions (first: %s, second: %s)", condition,
							firstMatchedSelector.getClass().getSimpleName(), knownSelector.getClass().getSimpleName()));
				}
				selectorVar = knownSelector.generate(condition, writer);
				assert selectorVar != null;
				firstMatchedSelector = knownSelector;
			}
		}

		if (selectorVar == null) {
			String message = String.format("Unsupported condition (found %s)", condition);
			LOGGER.log(Level.SEVERE, message);
			throw new IllegalArgumentException(message);
		}
		return selectorVar;
	}

	private static final <T extends Selector> String create(T element, StringWriter writer,
			List<com.is2t.giml.cssgenerator.selector.Selector> knownElements) {
		String selectorVar = null;
		com.is2t.giml.cssgenerator.selector.Selector firstMatchedSelector = null;
		for (com.is2t.giml.cssgenerator.selector.Selector knownSelector : knownElements) {
			if (knownSelector.matches(element)) {
				if (selectorVar != null) {
					throw new RuntimeException(String.format(
							"Selector '%s' matches multiple selector definitions (first: %s, second: %s)", element,
							firstMatchedSelector.getClass().getSimpleName(), knownSelector.getClass().getSimpleName()));
				}
				selectorVar = knownSelector.generate(element, writer);
				assert selectorVar != null;
				firstMatchedSelector = knownSelector;
			}
		}

		if (selectorVar == null) {
			String message = String.format("Unsupported selector (found %s)", element);
			LOGGER.log(Level.SEVERE, message);
			throw new IllegalArgumentException(message);
		}

		return selectorVar;
	}

}
