/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.generator.ant;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;

import com.is2t.giml.classpath.ClasspathModel;
import com.is2t.giml.classpath.processor.GIMLClasspathProcessor;
import com.is2t.giml.cssgenerator.CSSGenerator;
import com.is2t.giml.generator.GIMLGenerator;
import com.is2t.giml.generator.js.GlobalObjectGIML;
import com.is2t.giml.generator.js.GlobalObjectGIMLIdentifier;
import com.is2t.giml.jsgenerator.JavaScriptGenerator;
import com.is2t.giml.jsgenerator.nodes.InlinedCode;
import com.is2t.giml.model.GIMLModelFactory;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.parser.XMLParser;
import com.is2t.giml.xml.parser.XMLParserOptions;
import com.is2t.giml.xml.parser.utils.FileUtils;

/**
 * The class {@code GIMLGeneratorTask}
 */
public class GIMLGeneratorTask extends Task {

	private final XMLParserOptions _parserOptions; // NOSONAR

	private File _rootGIMLFile; // NOSONAR
	private File _moduleDirectory; // NOSONAR
	private final List<File> _moduleIncludeDirectories; // NOSONAR
	private File _outputDirectory; // NOSONAR
	private final List<File> _cssFiles; // NOSONAR
	private final List<String> _classpath; // NOSONAR

	/**
	 * Instantiate the task {@code GIMLGeneratorTask} with default parameters.
	 */
	public GIMLGeneratorTask() {
		super();

		this._parserOptions = new XMLParserOptions();
		this._cssFiles = new ArrayList<File>();
		this._classpath = new ArrayList<String>();
		this._moduleIncludeDirectories = new ArrayList<File>();
	}

	@Override
	public void execute() throws BuildException {
		super.execute(); // this is important

		// check root file and module directory exist
		if (!this.fileExists(this._rootGIMLFile)) {
			throw new BuildException(String.format("'%s' does not exist or does not denote a file path.",
					this._rootGIMLFile.getAbsolutePath()));
		}

		if(this._moduleDirectory != null){
			this._moduleIncludeDirectories.add(_moduleDirectory);
		}
		for(File moduleDirectory : _moduleIncludeDirectories){
			if (!this.directoryExists(moduleDirectory)) {
				throw new BuildException(String.format("'%s' does not exist or does not denote a directory path.",
						moduleDirectory.getAbsolutePath()));
			}
		}

		// Create output directory if necessary
		this._outputDirectory.mkdirs();
		if (!this._outputDirectory.exists()) {
			throw new BuildException(
					String.format("Cannot create directory '%s'", this._outputDirectory.getAbsolutePath()));
		}

		// Load classpath data
		ClasspathModel classpathModel = new GIMLClasspathProcessor()
				.process(_classpath.toArray(new String[_classpath.size()]));

		String javaScriptPackageName = JavaScriptGenerator.DEFAULT_GENERATED_PACKAGE_NAME;
		String javaScriptClassName = JavaScriptGenerator.DEFAULT_GENERATED_CLASS_NAME;

		// Parse XML files and generate sources
		InlinedCode[] inlinedCodes;
		GIMLGenerator gimlGenerator;
		try {
			XMLParser parser = new XMLParser(this._rootGIMLFile, this._moduleIncludeDirectories);
			parser.setOptions(this._parserOptions);
			Document rootDocument = parser.parse();
			rootDocument = GIMLModelFactory.createGIMLDocument(rootDocument);
			rootDocument = rootDocument.compact().resolveTemplates();
			gimlGenerator = new GIMLGenerator(rootDocument, this._outputDirectory, this._moduleDirectory);
			gimlGenerator.setJavaScriptClassFullyQualifiedName(javaScriptPackageName + '.' + javaScriptClassName);
			gimlGenerator.generate();
			inlinedCodes = gimlGenerator.getInlinedJavaScriptCodes();
		} catch (Exception e) {
			throw new BuildException(e);
		}

		// Parse CSS files and generate sources
		try {
			CSSGenerator generator = new CSSGenerator();
			for (File cssFile : this._cssFiles) {
				generator.parseFile(cssFile);
			}
			generator.generate(this._outputDirectory);
		} catch (Exception e) {
			throw new BuildException(e);
		}

		// Parse JavaScript files + inlined code and generate MicroEJ source code
		try {
			JavaScriptGenerator generator = new JavaScriptGenerator(classpathModel);
			generator.addInlinedCode(inlinedCodes);
			List<File> jsFiles = FileUtils.listFiles(_moduleDirectory, ".js");
			for (File jsFile : jsFiles) {
				generator.addJSFile(jsFile);
			}
			// populate global objects
			generator.addGlobalObject(new GlobalObjectGIML(gimlGenerator));
			Map<String, Integer> identifiers = gimlGenerator.getAllRootIndentifiers();
			for (String idName : identifiers.keySet()) {
				generator.addGlobalObject(new GlobalObjectGIMLIdentifier(idName, identifiers.get(idName)));
			}

			generator.generate(this._outputDirectory);
		} catch (Exception e) {
			throw new BuildException(e);
		}
	}

	/**
	 * Sets the root file of the described graphic interface.
	 *
	 * @param rootGIMLFile
	 *            the pathname denoting the root file of interface description.
	 */
	public void setRoot(File rootGIMLFile) {
		this._rootGIMLFile = rootGIMLFile;
	}

	/**
	 * Sets the directory where interface description modules are stored.
	 *
	 * @param moduleDirectory
	 *            the pathname denoting the directory where interface description modules are stored.
	 */
	public void setModuleDir(File moduleDirectory) {
		this._moduleDirectory = moduleDirectory;
	}

	/**
	 * Sets extra directories where interface description modules are stored.
	 *
	 * @param reference
	 *            a reference of the modules directories the pathname denoting
	 *            the directories where interface description modules are
	 *            stored.
	 */
	public void setModuleIncludeDirs(Reference reference) {
		Path cssFilesAsPath = new Path(getProject());
		cssFilesAsPath.setRefid(reference);
		for (String path : cssFilesAsPath.list()) {
			this._moduleIncludeDirectories.add(new File(path));
		}
	}

	/**
	 * Sets the generated source output directory.
	 *
	 * @param outputDirectory
	 *            the pathname denoting the directory where source are generated.
	 */
	public void setOutputDir(File outputDirectory) {
		this._outputDirectory = outputDirectory;
	}

	/**
	 * Sets the attribute name describing the module ID.
	 *
	 * @param moduleIDAttributeName
	 *            the attribute name describing the module ID.
	 */
	public void setModuleIdAttr(String moduleIDAttributeName) {
		this._parserOptions.setDocumentIdAttributeName(moduleIDAttributeName);
	}

	/**
	 * Sets the extension of module files.
	 *
	 * @param moduleFileExtension
	 *            the extension of module files.
	 */
	public void setModuleExt(String moduleFileExtension) {
		this._parserOptions.setXMLFileExtension(moduleFileExtension);
	}

	/**
	 * Sets the tag name describing module import.
	 *
	 * @param importTagName
	 *            the tag name describing module import.
	 */
	public void setImportTag(String importTagName) {
		this._parserOptions.setImportTagName(importTagName);
	}

	/**
	 * Sets the attribute name describing module ID to import.
	 *
	 * @param importIDAttributeName
	 *            the attribute name describing module ID to import.
	 */
	public void setImportIdAttr(String importIDAttributeName) {
		this._parserOptions.setImportTagAttributeName(importIDAttributeName);
	}

	/**
	 * Sets the tag name describing template declaration.
	 *
	 * @param templateTagName
	 *            the tag name describing template declaration.
	 */
	public void setTemplateTag(String templateTagName) {
		this._parserOptions.setTemplateTagName(templateTagName);
	}

	/**
	 * Sets the attribute name describing template ID.
	 *
	 * @param templateIDAttributeName
	 *            the attribute name describing template ID.
	 */
	public void setTemplateIdAttr(String templateIDAttributeName) {
		this._parserOptions.setTemplateTagAttributeName(templateIDAttributeName);
	}

	/**
	 * Sets the prefix to use to identify template parameter beginning.
	 *
	 * @param templateParameterPrefix
	 *            the prefix to use to identify template parameter beginning.
	 */
	public void setTemplateParameterPrefix(String templateParameterPrefix) {
		this._parserOptions.setTemplateParameterPrefix(templateParameterPrefix);
	}

	/**
	 * Sets the suffix to use to identify template parameter ending.
	 *
	 * @param templateParameterSuffix
	 *            the suffix to use to identify template parameter ending.
	 */
	public void setTemplateParameterSuffix(String templateParameterSuffix) {
		this._parserOptions.setTemplateParameterSuffix(templateParameterSuffix);
	}

	/**
	 * Sets the tag name describing template content declaration.
	 *
	 * @param templateContentTag
	 *            the tag name describing template content declaration.
	 */
	public void setTemplateContentTag(String templateContentTag) {
		this._parserOptions.setTemplateContentTagName(templateContentTag);
	}

	/**
	 * Sets the attribute name describing template content ID during template declaration.
	 *
	 * @param templateContentIDAttributeName
	 *            the attribute name describing template content ID during template declaration.
	 */
	public void setTemplateContentIdAttr(String templateContentIDAttributeName) {
		this._parserOptions.setTemplateContentTagAttributeName(templateContentIDAttributeName);
	}

	/**
	 * Sets the attribute name describing template content reference ID during template instantiation.
	 *
	 * @param templateContentInstantiationIDAttributeName
	 *            the attribute name describing template content reference ID during template instantiation.
	 */
	public void setTemplateContentInstantiationIdAttr(String templateContentInstantiationIDAttributeName) {
		this._parserOptions
		.setTemplateContentTagInstantiationAttributeName(templateContentInstantiationIDAttributeName);
	}

	/**
	 * Sets CSS files to parse.
	 *
	 * @param reference
	 *            a reference to CSS files to parse.
	 */
	public void setCSSFilesRef(Reference reference) {
		Path cssFilesAsPath = new Path(getProject());
		cssFilesAsPath.setRefid(reference);
		for (String path : cssFilesAsPath.list()) {
			this._cssFiles.add(new File(path));
		}
	}

	public void setClasspathRef(Reference reference) {
		Path cpFilesAsPath = new Path(getProject());
		cpFilesAsPath.setRefid(reference);
		for (String path : cpFilesAsPath.list()) {
			this._classpath.add(path);
		}
	}

	private boolean fileExists(File file) {
		return file.exists() && file.isFile();
	}

	private boolean directoryExists(File dir) {
		return dir.exists() && dir.isDirectory();
	}

	private void deleteDirectoriesAndFiles(File basedir) throws IOException {
		try {
			if (basedir.exists()) {
				File[] files = basedir.listFiles();
				if (files == null) {
					return;
				}
				for (File file : files) {
					if (file.isDirectory()) {
						deleteDirectoriesAndFiles(file);
					}
					if (!file.delete()) {
						throw new IOException(String.format("Cannot delete file '%s'.", file.getAbsolutePath()));
					}
				}
				if (!basedir.delete()) {
					throw new IOException(String.format("Cannot delete directory '%s'.", basedir.getAbsolutePath()));
				}
			}
		} catch (IOException e) {
			// we don't care, just for test!
		}
	}

}
