/*
 * Java
 *
 * Copyright 2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.jsgenerator;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;
import org.stringtemplate.v4.StringRenderer;

import com.is2t.giml.classpath.ClasspathModel;
import com.is2t.giml.jsgenerator.global.GlobalObject;
import com.is2t.giml.jsgenerator.global.GlobalObjectConsole;
import com.is2t.giml.jsgenerator.global.GlobalObjectFunction;
import com.is2t.giml.jsgenerator.nodes.GeneratedMethod;
import com.is2t.giml.jsgenerator.nodes.InlinedCode;
import com.is2t.java.JavaSourceToStringGenerator;
import com.is2t.java.nodes.Expression;
import com.is2t.java.nodes.MethodDeclaration;

import jdk.nashorn.internal.ir.FunctionNode;
import jdk.nashorn.internal.ir.IdentNode;
import jdk.nashorn.internal.ir.Statement;


/**
 * The class {@code JavaScriptGenerator} generates all GIML JavaScript code to
 * MicroEJ source code.
 */
public class JavaScriptGenerator {

	/**
	 * Default generated package name.
	 */
	public static final String DEFAULT_GENERATED_PACKAGE_NAME = "com.is2t.giml.js";

	/**
	 * Default generated class name.
	 */
	public static final String DEFAULT_GENERATED_CLASS_NAME = "GIMLJavaScriptCode";

	private static final Logger LOGGER = Logger.getLogger(JavaScriptGenerator.class.getSimpleName());

	private static final String TEMPLATE_PATH = "com/is2t/giml/jsgenerator/template.stg";
	private static final String TEMPLATE_GROUP = "JavaScriptGenerator";
	private static final String TEMPLATE_PACKAGE_VAR = "package";
	private static final String TEMPLATE_CLASS_VAR = "class";
	private static final String TEMPLATE_METHOD_VAR = "method";

	private String _generatedPackageName; // NOSONAR
	private String _generatedClassName; // NOSONAR

	private final List<InlinedCode> _inlinedCodeFunctions;
	private final List<File> _jsFiles;

	private final ArrayList<GeneratedMethod> generatedMethods;

	private final HashMap<String, GlobalObject> globalObjects;

	private final ClasspathModel classpathModel;

	/**
	 * Instantiate a new {@code CSSGenerator} with default values.
	 *
	 * <ul>
	 * <li>Default generated package name is
	 * {@value JavaScriptGenerator#DEFAULT_GENERATED_PACKAGE_NAME}</li>
	 * <li>Default generated class name is
	 * {@value JavaScriptGenerator#DEFAULT_GENERATED_CLASS_NAME}</li>
	 * <li>Default generated method name is
	 * {@value JavaScriptGenerator#DEFAULT_GENERATED_METHOD_NAME}</li>
	 * </ul>
	 *
	 * @param classpathModel
	 */
	public JavaScriptGenerator(ClasspathModel classpathModel) {
		this.classpathModel = classpathModel;
		this._generatedPackageName = DEFAULT_GENERATED_PACKAGE_NAME;
		this._generatedClassName = DEFAULT_GENERATED_CLASS_NAME;
		this._inlinedCodeFunctions = new ArrayList<>();
		this._jsFiles = new ArrayList<>();
		this.generatedMethods = new ArrayList<>();
		globalObjects = new HashMap<>();
		addGlobalObject(new GlobalObjectConsole());
	}

	/**
	 * Generates all saved CSS rules into a Java source code file.
	 *
	 * @param outputDir
	 *            the output directory where the generated Java class will be generated. This directory must be a source
	 *            directory. All needed packages will be automatically generated.
	 * @throws IllegalArgumentException
	 *             if given output directory denotes a file path.
	 * @throws IOException
	 *             if something went wrong during file generation.
	 *
	 * @see #parseFile(File)
	 * @see #setGeneratedPackageName(String)
	 * @see #setGeneratedClassName(String)
	 */
	public void generate(File outputDir) throws IllegalArgumentException, IOException {
		if (outputDir.exists() && outputDir.isFile()) {
			throw new IllegalArgumentException(
					String.format("%s does not denote a directory path.", outputDir.getAbsolutePath()));
		}

		if (!outputDir.exists()) {
			outputDir.mkdirs();
		}

		// Generate JavaScript code to MicroEJ code
		JavaScriptParser jsParser = new JavaScriptParser();
		JavaScriptSourcePreGenerator jsSourcePreGenerator = new JavaScriptSourcePreGenerator();
		final JavaScriptSourceToMicroEJSourceGenerator jsSourceToJavaSourceGenerator = new JavaScriptSourceToMicroEJSourceGenerator(
				this, jsSourcePreGenerator);
		JavaSourceToStringGenerator javaSourceToStringGenerator = new JavaSourceToStringGenerator();
		javaSourceToStringGenerator
		.setCurrentTypeReference(jsSourceToJavaSourceGenerator.getTypeReference(getGeneratedClassname()));
		List<Statement> initCodeStatements = new ArrayList<>();

		// generate code inlined in GIML attributes
		for (InlinedCode inlinedCodeFunction : _inlinedCodeFunctions) {
			String jsCode = inlinedCodeFunction.getSource();
			List<Statement> statements;
			try {
				statements = jsParser.parseInlinedCode(new StringReader(jsCode));
			} catch (IOException e) {
				throw new AssertionError(e);
			}
			if (statements != null) {
				generatedMethods.add(generatedMethod(jsSourceToJavaSourceGenerator, javaSourceToStringGenerator,
						inlinedCodeFunction.getFunctionName(), statements));
			}
			// else: parse error have been dumped to the standard output
		}

		// generate code declared in .js file
		for (File jsFile : _jsFiles) {
			List<Statement> statements;
			try (FileReader fileReader = new FileReader(jsFile)) {
				statements = jsParser.parseInlinedCode(fileReader);
			}
			if (statements != null) {
				initCodeStatements.addAll(statements);
			}
			// else: parse error have been dumped to the standard output
		}

		// one global pass before generation
		for (Statement statement : initCodeStatements) {
			statement.accept(jsSourcePreGenerator);
		}

		// generate the global init function statements with global object
		// initialization (the JS World)
		MethodGenerator initJSSpecificGen = new MethodGenerator() {

			@Override
			public void postGen(List<com.is2t.java.nodes.Statement> statements) {
				// generate and insert global objects initialization
				Map<String, GlobalObject> globalObjects = getGlobalObjects();
				List<com.is2t.java.nodes.Statement> initJavaStatements = new ArrayList<>();
				for (GlobalObject globalObject : globalObjects.values()) {
					initJavaStatements.add(globalObject.generateInitialization(jsSourceToJavaSourceGenerator));
				}

				statements.addAll(0, initJavaStatements);
			}
		};

		// All statements generated => generate function statements
		for (FunctionNode declaredFunction : jsSourcePreGenerator.getFunctions()) {
			MethodDeclaration method = jsSourceToJavaSourceGenerator.generate(declaredFunction);
			String generatedJavaCode = javaSourceToStringGenerator.toString(method);

			generatedMethods.add(new GeneratedMethod(generatedJavaCode));
		}

		GeneratedMethod initMethod = generatedMethod(jsSourceToJavaSourceGenerator, javaSourceToStringGenerator,
				"initJS", initJSSpecificGen, initCodeStatements);
		generatedMethods.add(initMethod);

		for (InstanceMethodCall call : jsSourceToJavaSourceGenerator.getInstanceMethodCalls()) {
			MethodDeclaration method = jsSourceToJavaSourceGenerator.generateInstanceMethodCall(classpathModel, call);
			String generatedJavaCode = javaSourceToStringGenerator.toString(method);

			generatedMethods.add(new GeneratedMethod(generatedJavaCode));
		}

		// Generate Java source code file
		this.generateJSFile(outputDir);
	}

	private GeneratedMethod generatedMethod(JavaScriptSourceToMicroEJSourceGenerator jsSourceToJavaSourceGenerator,
			JavaSourceToStringGenerator javaSourceToStringGenerator, String functionName, List<Statement> javaScriptStatements) {
		return generatedMethod(jsSourceToJavaSourceGenerator, javaSourceToStringGenerator, functionName,
				null,
				javaScriptStatements);
	}

	private static interface MethodGenerator {
		public void postGen(List<com.is2t.java.nodes.Statement> statements);
	}

	private GeneratedMethod generatedMethod(JavaScriptSourceToMicroEJSourceGenerator jsSourceToJavaSourceGenerator,
			JavaSourceToStringGenerator javaSourceToStringGenerator, String functionName,
			MethodGenerator specificGen, List<Statement> javaScriptStatements) {
		StringBuilder javaCodeSb = new StringBuilder();

		List<com.is2t.java.nodes.Statement> allJavaStatements = new ArrayList<>();
		for (Statement statement : javaScriptStatements) {
			com.is2t.java.nodes.Statement javaStatement = jsSourceToJavaSourceGenerator.generate(statement);
			if (javaStatement != null) {
				allJavaStatements.add(javaStatement);
			}
		}
		jsSourceToJavaSourceGenerator.generateImplicitReturn(allJavaStatements);

		if (specificGen != null) {
			specificGen.postGen(allJavaStatements);
		}

		for (com.is2t.java.nodes.Statement javaStatement : allJavaStatements) {
			String generatedJavaCode = javaSourceToStringGenerator.toString(javaStatement);
			javaCodeSb.append(generatedJavaCode);
			if (javaStatement instanceof Expression) {
				javaCodeSb.append(';');
			}
			javaCodeSb.append("\n");
		}

		return new GeneratedMethod("public static Object " + functionName + "(){\n" + javaCodeSb.toString() + "}");
	}

	/**
	 * Sets the generated Java source file package name.
	 *
	 * @param name
	 *            the package name.
	 *
	 * @see #DEFAULT_GENERATED_PACKAGE_NAME
	 */
	public void setGeneratedPackageName(String name) {
		this._generatedPackageName = name;
	}

	/**
	 * Sets the generated Java source file class name.
	 *
	 * @param name
	 *            the class name.
	 *
	 * @see #DEFAULT_GENERATED_CLASS_NAME
	 */
	public void setGeneratedClassName(String name) {
		this._generatedClassName = name;
	}

	private void generateJSFile(File outputDir) throws IOException {
		STGroup stGroup = new STGroupFile(TEMPLATE_PATH);
		stGroup.registerRenderer(String.class, new StringRenderer());
		ST template = stGroup.getInstanceOf(TEMPLATE_GROUP);
		template.add(TEMPLATE_PACKAGE_VAR, this._generatedPackageName);
		template.add(TEMPLATE_CLASS_VAR, this._generatedClassName);

		for (GeneratedMethod gm : generatedMethods) {
			template.addAggr("methods.{content}", gm.getCodeContent());
		}
		for (GlobalObject go : getGlobalObjects().values()) {
			template.addAggr("globalObjects.{name}", go.getGlobalName());

			if(go instanceof GlobalObjectFunction) {
				GlobalObjectFunction func = (GlobalObjectFunction)go;
				template.addAggr("functionClasses.{name, args}", go.getGlobalName(),
						toStringArgs(func.getNode().getParameters()));
			}
		}

		File generatedFile = new File(outputDir, this.getGeneratedFileRelativePath());
		generatedFile.getParentFile().mkdirs();
		generatedFile.createNewFile();
		template.write(generatedFile, null);
	}

	private String toStringArgs(List<IdentNode> parameters) {
		StringBuilder sb = new StringBuilder();
		int argIndex = 0;
		for (IdentNode parameter : parameters) {
			if (argIndex > 0) {
				sb.append(',');
			}
			sb.append("objects[" + argIndex + "]");
			++argIndex;
		}
		return sb.toString();
	}

	private String getGeneratedFileRelativePath() {
		return this._generatedPackageName.replace('.', '/') + "/" + this._generatedClassName + ".java";
	}

	/**
	 * Add JavaScript pieces of code to be generated that have not been parsed from
	 * a .js file
	 *
	 * @param inlinedCodes
	 *            inline code functions to be generated to the MicroEJ source file
	 */
	public void addInlinedCode(InlinedCode[] inlinedCodes) {
		Collections.addAll(_inlinedCodeFunctions, inlinedCodes);
	}

	/**
	 * Add a new JavaScript file to be generated. All statements but functions will
	 * be generated in the global "init" method.
	 *
	 * @param jsFile
	 *            a <code>.js</file)
	 */
	public void addJSFile(File jsFile) {
		_jsFiles.add(jsFile);
	}

	/**
	 * Register a well-known global object which is able to expand its functions
	 *
	 * @param globalObject
	 *            the global object to register
	 */
	public void addGlobalObject(GlobalObject globalObject) {
		String globalName = globalObject.getGlobalName();
		if (globalObjects.get(globalName) == null) {
			globalObjects.put(globalName, globalObject);
		}
	}

	/**
	 * Gets the registered global objects.
	 *
	 * @return the registered global objects.
	 */
	public Map<String, GlobalObject> getGlobalObjects() {
		return globalObjects;
	}

	/**
	 * @return the fully qualified name of the generated class
	 */
	public String getGeneratedClassname() {
		return _generatedPackageName + '.' + _generatedClassName;
	}
}
