/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.is2t.giml.description.Description;
import com.is2t.giml.description.DescriptionFactory;
import com.is2t.giml.model.exception.GIMLMisplacedElementException;
import com.is2t.giml.model.exception.GIMLTemplateInvocationException;
import com.is2t.giml.model.exception.GIMLUnknownElementException;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;

/**
 * The class {@code GIMLModelFactory} is a factory to create a GIML valid, normalized GIML model.
 *
 * @see #createGIMLDocument(Document)
 */
public class GIMLModelFactory {

	private GIMLModelFactory() {
		// Factory class
	}

	/**
	 * Creates a valid and normalized GIML model from the given document.
	 *
	 * A model is valid if:
	 * <ul>
	 * <li>no template has the same name as a GIML element name,</li>
	 * <li>no element is misplaced.</li>
	 * </ul>
	 *
	 * A model is normalized if all optional elements are present is the model.
	 *
	 * @param document
	 *            the document to transform into a GIML model.
	 * @return a GIML valid and normalized document
	 * @throws GIMLTemplateInvocationException
	 *             if a template has the same name as a GIML element.
	 * @throws GIMLUnknownElementException
	 *             if an document element is unknown for GIML model.
	 * @throws GIMLMisplacedElementException
	 *             if an element is misplaced in the model.
	 */
	public static Document createGIMLDocument(Document document)
			throws GIMLTemplateInvocationException, GIMLUnknownElementException, GIMLMisplacedElementException {
		checkTemplateUsage(document);
		checkMisplacedElements(document);
		return normalizeDocument(document);
	}

	private static void checkTemplateUsage(Document document) throws GIMLTemplateInvocationException {
		// Work on a compacted document since it is a lot easier to process
		document = document.compact();

		// Loop over exists tags, check if there exists a template that have the same name as a GIML object.
		Set<String> allTags = DescriptionFactory.getAllTags();
		for (Template template : document.getTemplates()) {
			String templateTagName = template.getName();
			if (allTags.contains(templateTagName)) {
				// Tag already exists. It is not safe to continue model creation, since it is impossible to
				// differentiate a template "xxx" with attribute "aaa" and a GIML object "xxx" with attribute "aaa"
				Description description = DescriptionFactory.getDescription(templateTagName);
				String errorMessage = String.format(
						"Cannot invoke template '%s' because a tag with the same name already exists", templateTagName);
				if (description != null) {
					String className = description.getElementType(templateTagName).getTypeName();
					StringBuilder builder = new StringBuilder();
					builder.append(errorMessage);
					builder.append(String.format(" (corresponding to class '%s')", className));
					errorMessage = builder.toString();
				}
				throw new GIMLTemplateInvocationException(errorMessage);
			}
		}
	}

	private static void checkMisplacedElements(Document document)
			throws GIMLUnknownElementException, GIMLMisplacedElementException {
		// Work on a compacted document since it is a lot easier to process
		// Work on a resolved document because to check node position, we need to replace templates by actual nodes.
		document = document.compact().resolveTemplates();

		for (Node documentChild : document.getChildren()) {
			checkMisplacedNode(document, documentChild, true);
		}
	}

	private static void checkMisplacedNode(Node parent, Node node, boolean isRoot)
			throws GIMLUnknownElementException, GIMLMisplacedElementException {
		String childTagName = node.getName();
		Description description = DescriptionFactory.getDescription(childTagName);

		checkElementExists(parent, childTagName);
		if (description != null && isRoot && !description.isRoot()) {
			throw new GIMLMisplacedElementException(
					String.format("Element '%s' cannot be a direct child of module root element", childTagName));
		}
		if (description != null && !isRoot && description.isRoot()) {
			throw new GIMLMisplacedElementException(
					String.format("Element '%s' must be a direct child of module root element", childTagName));
		}

		for (Node child : node.getChildren()) {
			checkMisplacedNode(node, child, false); // obviously children cannot be root elements
		}
	}

	private static void checkElementExists(Node parent, String childTagName) throws GIMLUnknownElementException {
		Description description = DescriptionFactory.getDescription(childTagName);
		if (description == null) {
			// Maybe it's an element that only exists in the parent element
			Description parentDescription = DescriptionFactory.getDescription(parent.getName());
			if (parentDescription == null || !parentDescription.getElements().contains(childTagName)) {
				// Tag is not a specific parent element, so this element is unknown for sure
				throw new GIMLUnknownElementException(
						String.format("Cannot find element '%s' in model description", childTagName));
			}
		}
	}

	private static Document normalizeDocument(Document document) throws NullPointerException, IllegalArgumentException {
		// Normalize children
		List<Node> children = new ArrayList<Node>();
		for (Node child : document.getChildren()) {
			children.add(normalizeNode(child));
		}

		// Normalize template root node
		List<Template> templates = new ArrayList<Template>();
		for (Template template : document.getTemplates()) {
			templates.add(normalizeTemplate(template));
		}

		// Normalize referenced documents
		List<Document> referencedDocuments = new ArrayList<Document>();
		for (Document refDocument : document.getReferencedDocuments()) {
			referencedDocuments.add(normalizeDocument(refDocument));
		}

		// Recreate a new document
		return new Document(document.getId(), document.getName(), document.getAttributes(), children,
				referencedDocuments, templates);
	}

	private static Node normalizeNode(Node node) throws IllegalArgumentException {
		Description nodeDescription = DescriptionFactory.getDescription(node.getName());
		String nodeDefaultElement = null;
		if (nodeDescription != null) {
			nodeDefaultElement = nodeDescription.getDefaultElement();
		}

		// Normalize node children first
		List<Node> children = new ArrayList<Node>();
		for (Node child : node.getChildren()) {
			if (nodeDefaultElement == null || nodeDefaultElement.equals(child.getName())) {
				// There is no default element or the default element is already present. So there is nothing to
				// normalize
				children.add(normalizeNode(child));
			} else {
				// There exists a default element AND this child is not the default element
				// So a default element is created, and set this child as the default element child. The newly created
				// default element has no attribute (since it is optional)
				List<Node> childChildren = new ArrayList<Node>();
				childChildren.add(normalizeNode(child));
				Node childDefaultNode = new Node(nodeDefaultElement, new HashSet<Attribute>(), childChildren);
				children.add(childDefaultNode);
			}
		}

		return new Node(node.getName(), node.getAttributes(), children);
	}

	private static Template normalizeTemplate(Template template) throws NullPointerException, IllegalArgumentException {
		// Normalizing a template is equivalent to normalizing its root node
		List<Node> children = new ArrayList<Node>();
		for (Node child : template.getTemplateNode().getChildren()) {
			children.add(normalizeNode(child));
		}
		Node templateNode = new Node(template.getTemplateTagName(), template.getTemplateNode().getAttributes(),
				children);

		return new Template(templateNode, template.getTemplateTagName(), template.getTemplateTagAttributeName(),
				template.getParameterPrefix(), template.getParameterSuffix(), template.getContentTagName(),
				template.getContentTagAttributeName(), template.getContentTagInstantiationAttributeName());
	}
}
