/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

/**
 * The <code>Document</code> class is a node container. It represents a tree-like structure split in nodes.
 *
 * @see Node
 */
public class Document extends Node {

	/**
	 * The unique ID of this document.
	 */
	protected String _id; // NOSONAR

	/**
	 * The referenced document inside this document.
	 */
	protected Set<Document> _referencedDocuments; // NOSONAR

	/**
	 * List of templates declared in this document.
	 */
	protected Map<String, Template> _templates; // NOSONAR

	/**
	 * Create an empty document with the given name.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the name of this document.
	 * @throws IllegalArgumentException
	 *             if given name or id are empty.
	 */
	public Document(String id, String name) throws IllegalArgumentException {
		super(name);
		this.initDocument(id);
	}

	/**
	 * Create a new node with given name and attributes.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the name of this document.
	 * @param attributes
	 *            declared attributes of the document to create.
	 * @throws IllegalArgumentException
	 *             if given name or id are empty, or if given name, id or attribute set are <code>null</code>.
	 */
	public Document(String id, String name, Set<Attribute> attributes) throws IllegalArgumentException {
		super(name, attributes);
		this.initDocument(id);
	}

	/**
	 * Create a new document with given name, attributes and children.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the name of this document.
	 * @param attributes
	 *            declared attributes of the document to create.
	 * @param children
	 *            children of this document.
	 * @throws IllegalArgumentException
	 *             if given name or id are empty, or if given name, id, attribute set or child list are
	 *             <code>null</code>.
	 */
	public Document(String id, String name, Set<Attribute> attributes, List<Node> children)
			throws IllegalArgumentException {
		super(name, attributes, children);
		this.initDocument(id);
	}

	/**
	 * Create a new document with given name, attributes and children.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the name of this document.
	 * @param attributes
	 *            declared attributes of the document to create.
	 * @param children
	 *            children of this document.
	 * @param referencedDocuments
	 *            documents that are referenced in the document to create.
	 * @throws IllegalArgumentException
	 *             if given name or id are empty, or if given name, id, attribute set, child list or referenced document
	 *             list are <code>null</code>.
	 */
	public Document(String id, String name, Set<Attribute> attributes, List<Node> children,
			List<Document> referencedDocuments) throws IllegalArgumentException {
		super(name, attributes, children);
		this.initDocument(id, referencedDocuments);
	}

	/**
	 * Create a new document with given name, attributes and children.
	 *
	 * @param id
	 *            the unique ID of this document.
	 * @param name
	 *            the name of this document.
	 * @param attributes
	 *            declared attributes of the document to create.
	 * @param children
	 *            children of this document.
	 * @param referencedDocuments
	 *            documents that are referenced in the document to create.
	 * @param templates
	 *            list of all declared templates in the document to create.
	 * @throws IllegalArgumentException
	 *             if given name or id are empty, or if given name, id, attribute set, child list, referenced document
	 *             list or template list are <code>null</code>.
	 */
	public Document(String id, String name, Set<Attribute> attributes, List<Node> children,
			List<Document> referencedDocuments, List<Template> templates) throws IllegalArgumentException {
		super(name, attributes, children);
		this.initDocument(id, referencedDocuments, templates);
	}

	/**
	 * Gets the unique ID of this document.
	 *
	 * @return the unique ID of this document.
	 */
	public String getId() {
		return this._id;
	}

	/**
	 * Gets the number of referenced documents in this document.
	 *
	 * @return the number of referenced documents in this document.
	 */
	public final int getReferencedDocumentCount() {
		return this._referencedDocuments.size();
	}

	/**
	 * <p>
	 * Gets a read-only list of referenced documents in this document.
	 * </p>
	 * <p>
	 * Any attempt to modify this list will result in a {@link UnsupportedOperationException}.
	 * </p>
	 *
	 * @return a possibly empty list of referenced documents.
	 */
	public List<Document> getReferencedDocuments() {
		List<Document> documents = new ArrayList<Document>();
		documents.addAll(this._referencedDocuments);
		return Collections.unmodifiableList(documents);
	}

	/**
	 * Gets the number of templates in this document.
	 *
	 * @return the number of templates in this document.
	 */
	public final int getTemplateCount() {
		return this._templates.size();
	}

	/**
	 * Gets the template with the given name declared in this document.
	 *
	 * @param templateName
	 *            the name of the template to find.
	 * @return the template, or <code>null</code> if no template with the given name exists in this document.
	 */
	public Template getTemplate(String templateName) {
		return this._templates.get(templateName);
	}

	/**
	 * <p>
	 * Gets a read-only list of templates in this document.
	 * </p>
	 * <p>
	 * Any attempt to modify this list will result in a {@link UnsupportedOperationException}.
	 * </p>
	 *
	 * @return a possibly empty list of templates.
	 */
	public List<Template> getTemplates() {
		List<Template> templates = new ArrayList<Template>();
		templates.addAll(this._templates.values());
		return Collections.unmodifiableList(templates);
	}

	/**
	 * Create a copy of this document containing content of the original document and all its referenced documents. As a
	 * consequence, a compacted document does not have referenced documents.
	 *
	 * @return a compacted document.
	 *
	 * @throws IllegalArgumentException
	 *             if a copy of this document cannot be created due to invalid parameters.
	 */
	public Document compact() throws IllegalArgumentException {
		return compact(new ArrayList<Document>());
	}

	protected Document compact(List<Document> browsedDocuments) {
		browsedDocuments.add(this);
		List<Node> allNodes = new ArrayList<Node>();
		List<Template> templates = new ArrayList<Template>();
		for (Document referencedDocument : this._referencedDocuments) {
			if (!browsedDocuments.contains(referencedDocument)) {
				Document compactedDocument = referencedDocument.compact(browsedDocuments);
				allNodes.addAll(compactedDocument.getChildren());
				templates.addAll(compactedDocument.getTemplates());
			}
		}
		allNodes.addAll(this._children);
		templates.addAll(this.getTemplates());
		return new Document(this._id, this._name, this._attributes, allNodes, new ArrayList<Document>(), templates);
	}

	/**
	 * Creates a copy of this document with all templates resolved. As a consequence, the returned document does not
	 * have templates anymore.
	 *
	 * @return a copy of this document with resolved templates.
	 *
	 * @throws IllegalArgumentException
	 *             if a copy of this document cannot be created due to invalid parameters.
	 */
	public Document resolveTemplates() throws IllegalArgumentException {
		List<Document> referencedDocuments = new ArrayList<Document>();

		// Resolve templates in referenced documents
		for (Document referenceDocument : this._referencedDocuments) {
			referencedDocuments.add(referenceDocument.resolveTemplates());
		}

		// Resolve templates in child nodes
		List<Node> previousResolve = new ArrayList<Node>();
		List<Node> resolvedNodes = this.resolve(this._children);

		do {
			previousResolve.clear();
			previousResolve.addAll(resolvedNodes);
			resolvedNodes = this.resolve(resolvedNodes);
		} while (!previousResolve.containsAll(resolvedNodes) || !resolvedNodes.containsAll(previousResolve));

		return new Document(this._id, this._name, this._attributes, resolvedNodes, referencedDocuments);
	}

	private List<Node> resolve(List<Node> children) {
		List<Node> resolvedChildren = new ArrayList<Node>();
		for (Node child : children) {
			List<Node> resolvedNodes = this.resolveTemplateInNode(child);
			if (resolvedNodes != null) {
				// resolved node is not null, so template was not empty.
				resolvedChildren.addAll(resolvedNodes);
			}
		}
		return resolvedChildren;
	}

	private List<Node> resolveTemplateInNode(Node node) throws IllegalArgumentException {
		// First, resolve template in node children
		List<Node> children = new ArrayList<Node>();
		for (Node child : node.getChildren()) {
			children.addAll(resolveTemplateInNode(child));
		}
		Node newNode = new Node(node.getName(), node.getAttributes(), children);

		// Then, check if this node is a template. If yes, replace it by an instantiation of the template
		Template template = this.getTemplateInTree(newNode.getName());
		if (template != null) {
			return template.instantiate(newNode);
		}

		List<Node> resolvedNodes = new ArrayList<Node>();
		resolvedNodes.add(newNode);
		return resolvedNodes;
	}

	private void initDocument(String id) throws IllegalArgumentException {
		if (id == null || id.isEmpty()) {
			throw new IllegalArgumentException("Document id cannot be null or empty");
		}
		this._id = id;
		this._referencedDocuments = new HashSet<Document>();
		this._templates = new HashMap<String, Template>();
	}

	private void initDocument(String id, List<Document> referencedDocuments) throws IllegalArgumentException {
		this.initDocument(id);
		if (referencedDocuments == null) {
			throw new IllegalArgumentException("Document referenced documents cannot be null");
		}
		this._referencedDocuments.addAll(referencedDocuments);
	}

	private void initDocument(String id, List<Document> referencedDocuments, List<Template> templates)
			throws IllegalArgumentException {
		this.initDocument(id, referencedDocuments);
		if (templates == null) {
			throw new IllegalArgumentException("Document templates cannot be null");
		}
		for (Template template : templates) {
			this._templates.put(template.getName(), template);
		}
	}

	private Template getTemplateInTree(String templateName) {
		// First, try to resolve in referenced documents
		for (Document includedDocument : this._referencedDocuments) {
			Template template = includedDocument.getTemplate(templateName);
			if (template != null) {
				return template;
			}
		}

		// Not in an included document, so must be in this document
		return this._templates.get(templateName);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof Document)) {
			return false;
		}

		Document other = (Document) obj;
		return this._id.equals(other._id) && this._name.equalsIgnoreCase(other._name)
				&& this._attributes.equals(other._attributes) && this._children.equals(other._children)
				&& this._referencedDocuments.equals(other._referencedDocuments)
				&& this._templates.equals(other._templates);
	}

	@Override
	public int hashCode() {
		return Objects.hash(this._id, this._name, this._attributes, this._children, this._referencedDocuments,
				this._templates);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		String parentToString = super.toString();

		builder.append("D[");
		builder.append("id=");
		builder.append(this._id);
		builder.append(',');
		builder.append(parentToString.subSequence(2, parentToString.length() - 1));

		if (this.getReferencedDocumentCount() > 0) {
			builder.append(',');
			builder.append("refDocuments=");
			builder.append('{');
			Iterator<Document> iterator = this._referencedDocuments.iterator();
			while (iterator.hasNext()) {
				builder.append(iterator.next().getId());
				if (iterator.hasNext()) {
					builder.append(',');
				}
			}
			builder.append('}');
		}
		if (this.getTemplateCount() > 0) {
			builder.append(',');
			builder.append("templates=");
			builder.append('{');
			Iterator<Template> iterator = this._templates.values().iterator();
			while (iterator.hasNext()) {
				builder.append(iterator.next().getName());
				if (iterator.hasNext()) {
					builder.append(',');
				}
			}
			builder.append('}');
		}
		builder.append(']');

		return builder.toString();
	}

}
