/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser.test.xmlparser;

import java.io.FileNotFoundException;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.parser.XMLParser;
import com.is2t.giml.xml.parser.XMLParserOptions;
import com.is2t.giml.xml.parser.exception.XMLParserException;

/**
 * Test cases of {@link XMLParser} on valid documents and resolved templates.
 */
public class XMLParserValidDocumentsResolvedTemplatesTest extends AbstractXMLParserValidTest {

	private static final String RESOURCE_FOLDER_PATH = "/com/is2t/giml/xml/parser/test/xmlparser/templates/valid/";

	private static final String PARSER_OPTION_XML_EXTENSION = ".xml";
	private static final String PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME = "name";
	private static final String PARSER_OPTION_IMPORT_TAGNAME = "import";
	private static final String PARSER_OPTION_IMPORT_ATTRIBUTE_NAME = "name";
	private static final boolean PARSER_OPTION_HANDLE_IMPORTS = true;
	private static final String PARSER_OPTION_TEMPLATE_TAGNAME = "template";
	private static final String PARSER_OPTION_TEMPLATE_ATTRIBUTE_NAME = "name";
	private static final String PARSER_OPTION_TEMPLATE_PARAMETER_PREFIX = "{{";
	private static final String PARSER_OPTION_TEMPLATE_PARAMETER_SUFFIX = "}}";
	private static final boolean PARSER_OPTION_HANDLE_TEMPLATES = true;
	private static final String PARSER_OPTION_TEMPLATE_CONTENT_TAG = "content";
	private static final String PARSER_OPTION_TEMPLATE_CONTENT_TAG_ATTRIBUTE_NAME = "id";
	private static final String PARSER_OPTION_TEMPLATE_CONTENT_TAG_INSTANTIATION_ATTRIBUTE_NAME = "ref";

	@Override
	protected XMLParserOptions getXMLParserOptions() {
		XMLParserOptions options = new XMLParserOptions();
		options.setXMLFileExtension(PARSER_OPTION_XML_EXTENSION);
		options.setDocumentIdAttributeName(PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME);
		options.setImportTagName(PARSER_OPTION_IMPORT_TAGNAME);
		options.setImportTagAttributeName(PARSER_OPTION_IMPORT_ATTRIBUTE_NAME);
		options.setHandleImports(PARSER_OPTION_HANDLE_IMPORTS);
		options.setTemplateTagName(PARSER_OPTION_TEMPLATE_TAGNAME);
		options.setTemplateTagAttributeName(PARSER_OPTION_TEMPLATE_ATTRIBUTE_NAME);
		options.setTemplateParameterPrefix(PARSER_OPTION_TEMPLATE_PARAMETER_PREFIX);
		options.setTemplateParameterSuffix(PARSER_OPTION_TEMPLATE_PARAMETER_SUFFIX);
		options.setTemplateContentTagName(PARSER_OPTION_TEMPLATE_CONTENT_TAG);
		options.setTemplateContentTagAttributeName(PARSER_OPTION_TEMPLATE_CONTENT_TAG_ATTRIBUTE_NAME);
		options.setTemplateContentTagInstantiationAttributeName(
				PARSER_OPTION_TEMPLATE_CONTENT_TAG_INSTANTIATION_ATTRIBUTE_NAME);
		options.setHandleTemplates(PARSER_OPTION_HANDLE_TEMPLATES);
		return options;
	}

	@Override
	protected String getResourceFolder() {
		return RESOURCE_FOLDER_PATH;
	}

	@Test
	public void testSimpleTemplate() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("simple/simple.xml")).parse();

		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "simple");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check before resolving template
		Set<Attribute> invocationNodeAttributes = createAttributeSet("name", "John");
		Node invocationNode = createNode("my-template", invocationNodeAttributes);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 1);
		Assert.assertTrue(document.getChildren().contains(invocationNode));

		// Check after resolving template
		document = document.resolveTemplates();
		Set<Attribute> resolveNodeAttributes = createAttributeSet("text", "Hello John!");
		Node resolveNode = createNode("node1", resolveNodeAttributes);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(resolveNode));
	}

	@Test
	public void testMultipleTemplatesWithChildrens() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("multiple/multiple.xml")).parse();

		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "multiple");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check before resolving template
		Set<Attribute> invocationNode1Attributes = createAttributeSet("name", "John");
		Node invocationNode1 = createNode("say-hello", invocationNode1Attributes);
		Node invocationNode2Child = createNode(PARSER_OPTION_TEMPLATE_CONTENT_TAG,
				createAttributeSet(PARSER_OPTION_TEMPLATE_CONTENT_TAG_INSTANTIATION_ATTRIBUTE_NAME, "myContent"),
				createNode("node2", NO_ATTRIBUTE));
		Set<Attribute> invocationNode2Attributes = createAttributeSet("firstname", "John", "lastname", "Doe");
		Node invocationNode2 = createNode("say-goodbye", invocationNode2Attributes, invocationNode2Child);
		Node container = createNode("node", NO_ATTRIBUTE, invocationNode1, invocationNode2);

		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 2);
		Assert.assertTrue(document.getChildren().contains(container));

		// Check after resolving template
		document = document.resolveTemplates();
		Node expectedNode = createNode("node", NO_ATTRIBUTE,
				createNode("node1", createAttributeSet("text", "Hello John!")), createNode("node1",
						createAttributeSet("text", "Goodbye John Doe!"), createNode("node2", NO_ATTRIBUTE)));
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(expectedNode));
	}

	@Test
	public void testSimpleTemplateInAReferencedDocument() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("withimport/doc1.xml")).parse();

		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "doc1");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getReferencedDocumentCount(), 1);

		// Check before resolving template
		Set<Attribute> invocationNodeAttributes = createAttributeSet("name", "John");
		Node invocationNode = createNode("my-template", invocationNodeAttributes);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(invocationNode));

		// Check after resolving template
		document = document.resolveTemplates();
		Set<Attribute> resolveNodeAttributes = createAttributeSet("text", "Hello John!");
		Node resolveNode = createNode("node1", resolveNodeAttributes);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(resolveNode));
	}

	@Test
	public void testNestedTemplates() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("nested/nested.xml")).parse();

		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "nested");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check before resolving template
		Set<Attribute> invocationAttributesT1 = createAttributeSet("name", "Doe");
		Node invocationNode = createNode("templateRoot", invocationAttributesT1);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 2);
		Assert.assertTrue(document.getChildren().contains(invocationNode));

		// Check after resolving template
		document = document.resolveTemplates();
		Set<Attribute> resolveNodeAttributes = createAttributeSet("text", "Goodbye John Doe!");
		Node resolveNode = createNode("node1", resolveNodeAttributes);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertTrue(document.getChildren().contains(resolveNode));
	}

}
